@@ -34,11 +34,11 @@ import Data.List
 -- | Args specifies arguments to the QuickCheck driver
 data Args
   = Args
-  { replay     :: Maybe (StdGen,Int) -- ^ should we replay a previous test?
-  , maxSuccess :: Int                -- ^ maximum number of successful tests before succeeding
-  , maxDiscard :: Int                -- ^ maximum number of discarded tests before giving up
-  , maxSize    :: Int                -- ^ size to use for the biggest test cases
-  , chatty     :: Bool               -- ^ whether to print anything
+  { replay          :: Maybe (StdGen,Int) -- ^ should we replay a previous test?
+  , maxSuccess      :: Int                -- ^ maximum number of successful tests before succeeding
+  , maxDiscardRatio :: Int                -- ^ maximum number of discarded tests per successful test before giving up
+  , maxSize         :: Int                -- ^ size to use for the biggest test cases
+  , chatty          :: Bool               -- ^ whether to print anything
   }
  deriving ( Show, Read )
 
@@ -78,11 +78,11 @@ isSuccess _         = False
 -- | stdArgs are the default test arguments used
 stdArgs :: Args
 stdArgs = Args
-  { replay     = Nothing
-  , maxSuccess = 100
-  , maxDiscard = 500
-  , maxSize    = 100
-  , chatty     = True
+  { replay          = Nothing
+  , maxSuccess      = 100
+  , maxDiscardRatio = 10
+  , maxSize         = 100
+  , chatty          = True
 -- noShrinking flag?
   }
 
@@ -105,28 +105,30 @@ quickCheckWithResult a p =
      rnd <- case replay a of
               Nothing      -> newStdGen
               Just (rnd,_) -> return rnd
-     test MkState{ terminal          = tm
-                 , maxSuccessTests   = maxSuccess a
-                 , maxDiscardedTests = maxDiscard a
-                 , computeSize       = case replay a of
-                                         Nothing    -> computeSize'
-                                         Just (_,s) -> computeSize' `at0` s
-                 , numSuccessTests   = 0
-                 , numDiscardedTests = 0
-                 , collected         = []
-                 , expectedFailure   = False
-                 , randomSeed        = rnd
-                 , numSuccessShrinks = 0
-                 , numTryShrinks     = 0
+     test MkState{ terminal                  = tm
+                 , maxSuccessTests           = if exhaustive p then 1 else maxSuccess a
+                 , maxDiscardedTests         = if exhaustive p then maxDiscardRatio a else maxDiscardRatio a * maxSuccess a
+                 , computeSize               = case replay a of
+                                                 Nothing    -> computeSize'
+                                                 Just (_,s) -> computeSize' `at0` s
+                 , numSuccessTests           = 0
+                 , numDiscardedTests         = 0
+                 , numRecentlyDiscardedTests = 0
+                 , collected                 = []
+                 , expectedFailure           = False
+                 , randomSeed                = rnd
+                 , numSuccessShrinks         = 0
+                 , numTryShrinks             = 0
+                 , numTotTryShrinks          = 0
                  } (unGen (property p))
   where computeSize' n d
           -- e.g. with maxSuccess = 250, maxSize = 100, goes like this:
           -- 0, 1, 2, ..., 99, 0, 1, 2, ..., 99, 0, 2, 4, ..., 98.
           | n `roundTo` maxSize a + maxSize a <= maxSuccess a ||
             n >= maxSuccess a ||
-            maxSuccess a `mod` maxSize a == 0 = n `mod` maxSize a + d `div` 10
+            maxSuccess a `mod` maxSize a == 0 = (n `mod` maxSize a + d `div` 10) `min` maxSize a
           | otherwise =
-            (n `mod` maxSize a) * maxSize a `div` (maxSuccess a `mod` maxSize a) + d `div` 10
+            ((n `mod` maxSize a) * maxSize a `div` (maxSuccess a `mod` maxSize a) + d `div` 10) `min` maxSize a
         n `roundTo` m = (n `div` m) * m
         at0 f s 0 0 = s
         at0 f s n d = f n d
@@ -186,7 +188,7 @@ doneTesting st _f =
        return NoExpectedFailure{ labels = summary st,
                                  numTests = numSuccessTests st,
                                  output = theOutput }
-  
+
 giveUp :: State -> (StdGen -> Int -> Prop) -> IO Result
 giveUp st _f =
   do -- CALLBACK gave_up?
@@ -214,24 +216,31 @@ runATest st f =
                  ]
        ++ ")"
         )
-     let size = computeSize st (numSuccessTests st) (numDiscardedTests st)
+     let size = computeSize st (numSuccessTests st) (numRecentlyDiscardedTests st)
      MkRose res ts <- protectRose (reduceRose (unProp (f rnd1 size)))
      callbackPostTest st res
-     
+
+     let continue break st' | abort res = break st'
+                            | otherwise = test st'
+
      case res of
        MkResult{ok = Just True, stamp = stamp, expect = expect} -> -- successful test
-         do test st{ numSuccessTests = numSuccessTests st + 1
-                   , randomSeed      = rnd2
-                   , collected       = stamp : collected st
-                   , expectedFailure = expect
-                   } f
-       
+         do continue doneTesting
+              st{ numSuccessTests           = numSuccessTests st + 1
+                , numRecentlyDiscardedTests = 0
+                , randomSeed                = rnd2
+                , collected                 = stamp : collected st
+                , expectedFailure           = expect
+                } f
+
        MkResult{ok = Nothing, expect = expect} -> -- discarded test
-         do test st{ numDiscardedTests = numDiscardedTests st + 1
-                   , randomSeed        = rnd2
-                   , expectedFailure   = expect
-                   } f
-         
+         do continue giveUp
+              st{ numDiscardedTests         = numDiscardedTests st + 1
+                , numRecentlyDiscardedTests = numRecentlyDiscardedTests st + 1
+                , randomSeed                = rnd2
+                , expectedFailure           = expect
+                } f
+
        MkResult{ok = Just False} -> -- failed test
          do if expect res
               then putPart (terminal st) (bold "*** Failed! ")
@@ -288,7 +297,7 @@ success st =
               , let s' = [ t | (t,0) <- s ]
               , not (null s')
               ]
-  
+
   covers = [ ("only " ++ show occurP ++ "% " ++ fst (head lps) ++ "; not " ++ show reqP ++ "%")
            | lps <- groupBy first
                   . sort
@@ -301,8 +310,8 @@ success st =
                  reqP   = maximum (map snd lps)
            , occurP < reqP
            ]
-  
-  (x,_) `first` (y,_) = x == y 
+
+  (x,_) `first` (y,_) = x == y
 
   maxi = map (\lps -> (fst (head lps), maximum (map snd lps)))
        . groupBy first
@@ -321,7 +330,7 @@ localMin :: State -> P.Result -> [Rose P.Result] -> IO Int
 localMin st res _ | P.interrupted res = localMinFound st res
 localMin st res ts = do
   putTemp (terminal st)
-    ( short 26 (P.reason res)
+    ( short 26 (oneLine (P.reason res))
    ++ " (after " ++ number (numSuccessTests st+1) "test"
    ++ concat [ " and "
             ++ show (numSuccessShrinks st)
@@ -349,18 +358,26 @@ localMin' st res (t:ts) =
     callbackPostTest st res'
     if ok res' == Just False
       then foundFailure st{ numSuccessShrinks = numSuccessShrinks st + 1 } res' ts'
-      else localMin st{ numTryShrinks = numTryShrinks st + 1 } res ts
+      else localMin st{ numTryShrinks    = numTryShrinks st + 1,
+                        numTotTryShrinks = numTotTryShrinks st + 1 } res ts
 
 localMinFound :: State -> P.Result -> IO Int
 localMinFound st res =
-  do putLine (terminal st)
-       ( P.reason res
-      ++ " (after " ++ number (numSuccessTests st+1) "test"
-      ++ concat [ " and " ++ number (numSuccessShrinks st) "shrink"
-                | numSuccessShrinks st > 0
-                ]
-      ++ "):  "
-       )
+  do let report = concat [
+           "(after " ++ number (numSuccessTests st+1) "test",
+           concat [ " and " ++ number (numSuccessShrinks st) "shrink"
+                  | numSuccessShrinks st > 0
+                  ],
+           "): "
+           ]
+     if isOneLine (P.reason res)
+       then putLine (terminal st) (P.reason res ++ " " ++ report)
+       else do
+         putLine (terminal st) report
+         sequence_
+           [ putLine (terminal st) msg
+           | msg <- lines (P.reason res)
+           ]
      callbackPostFinalFailure st res
      return (numSuccessShrinks st)
 
