@@ -1,12 +1,14 @@
 module Test.QuickCheck.Text
   ( Str(..)
   , ranges
-  
+
   , number
   , short
   , showErr
+  , oneLine
+  , isOneLine
   , bold
-  
+
   , newTerminal
   , newStdioTerminal
   , newNullTerminal
@@ -62,6 +64,12 @@ short n s
 showErr :: Show a => a -> String
 showErr = unwords . words . show
 
+oneLine :: String -> String
+oneLine = unwords . words
+
+isOneLine :: String -> Bool
+isOneLine xs = '\n' `notElem` xs
+
 bold :: String -> String
 -- not portable:
 --bold s = "\ESC[1m" ++ s ++ "\ESC[0m"
@@ -129,7 +137,7 @@ putPart, putTemp, putLine :: Terminal -> String -> IO ()
 putPart tm@(MkTerminal _ out _) s =
   do flush tm
      put out s
-     
+
 putTemp tm@(MkTerminal _ _ err) s =
   do flush tm
      put err s
@@ -138,7 +146,7 @@ putTemp tm@(MkTerminal _ _ err) s =
        put err ( [ ' ' | _ <- s ]
               ++ [ '\b' | _ <- s ]
                )
-     
+
 putLine tm@(MkTerminal _ out _) s =
   do flush tm
      put out s
